

/* ***********************************************************************
                Naval Air Warfare Center / Aircraft Division              
                        Patuxent Naval Air Station                         
 ========================================================================

        File:           conversions.c                                         
                                                                        
        Classification: UNCLASSIFIED                                   
                                                                        
        Description:    This file contains the functions, methods, and
			global variables needed to perform unit
			conversions.
			
        Written by:	Alex Harper
*********************************************************************** */


#define FESD 1
#define FERD 2

#include <string.h>

#include "vissim.h"
#include "matrix.h"

#ifdef WIN32
#define fcos(xxx) cos(xxx)
#endif

double CONVERT_ctrlat;
double CONVERT_ctrlon;
double CONVERT_EarthRadius = EARTH_RADIUS;
double CONVERT_EarthRadius2 = EARTH_RADIUS*EARTH_RADIUS;
double CONVERT_sin_ctrlat;
double CONVERT_cos_ctrlat;
double CONVERT_sin_ctrlon;
double CONVERT_cos_ctrlon;

double CONVERT_sin_ctrlat_X_cos_ctrlon;
double CONVERT_sin_ctrlat_X_sin_ctrlon;
double CONVERT_cos_ctrlat_X_cos_ctrlon;
double CONVERT_cos_ctrlat_X_sin_ctrlon;


/*****************************************************************************
   deg_to_dms
   Convert latitude and longitude in decimal degrees to deg:min:sec
   Inputs : d_lat - latitude
            d_lon - longitude
            lat - character string address for latitude
            lon - character string address for longitude
            format
   Outputs : none
*****************************************************************************/
void deg_to_dms(double d_lat, double d_lon, char *lat, char *lon, int format) {

   int     deg1, min1, sec1;
   int     deg2, min2, sec2;
   double  fsec1, fsec2;
   char    dir1, dir2;

   if (d_lat<0.0) {
      dir1='S';
      d_lat=-d_lat;
   }

   else
      dir1='N';

   deg1=(int)d_lat;
   min1=(int)((d_lat-(double)deg1)*60.0);
   sec1=(int)((d_lat-(double)deg1-(double)min1/60.0)*3600.0);
   fsec1=(int)((d_lat-(double)deg1-(double)min1/60.0)*3600.0);

   if (d_lon<0.0) { dir2='W'; d_lon=-d_lon; }
   else { dir2='E'; }
   deg2=(int)d_lon;
   min2=(int)((d_lon-(double)deg2)*60.0);
   sec2=(int)((d_lon-(double)deg2-(double)min2/60.0)*3600.0);
   fsec2=(int)((d_lon-(double)deg2-(double)min2/60.0)*3600.0);

   switch (format) {
   case 1:
      sprintf(lat,"%2.2d:%2.2d:%2.2d%c",deg1,min1,sec1,dir1);
      sprintf(lon,"%3.3d:%2.2d:%2.2d%c",deg2,min2,sec2,dir2);
      break;
   case 2:
      sprintf(lat,"%2.2d:%2.2d:%05.2lf%c",deg1,min1,fsec1,dir1);
      sprintf(lon,"%3.3d:%2.2d:%05.2lf%c",deg2,min2,fsec2,dir2);
      break;
   case 3:
      sprintf(lat,"%2.2d%2.2d%2.2d%c",deg1,min1,sec1,dir1);
      sprintf(lon,"%3.3d%2.2d%2.2d%c",deg2,min2,sec2,dir2);
      break;
   case 4:
      sprintf(lat,"%2.2d%2.2d%05.2lf%c",deg1,min1,fsec1,dir1);
      sprintf(lon,"%3.3d%2.2d%05.2lf%c",deg2,min2,fsec2,dir2);
      break;
   default:
      lat[0]=0;
      lon[0]=0;
      break;
   }
}


/*****************************************************************************
   dms_to_deg
   Convert latitude and longitude deg:min:sec to decimal degrees
   Inputs : lat - latitude character string
            lon - longitude character string
            d_lat - latitude
            d_lon - longitude
   Outputs : none
*****************************************************************************/
void dms_to_deg(char *lat, char *lon, double *d_lat, double *d_lon) {

   double  deg, min, sec;
   char    dir;
 
   deg = min = sec = 0.0;
   dir = lat[strlen(lat)-1];
   sscanf(lat,"%lf:%lf:%lf",&deg,&min,&sec);
   *d_lat = deg + min/60.0 + sec/3600.0;
   if (dir=='S' || dir=='s') *d_lat *= -1.0;

   deg = min = sec = 0.0;
   dir = lon[strlen(lon)-1];
   sscanf(lon,"%lf:%lf:%lf",&deg,&min,&sec);
   *d_lon = deg + min/60.0 + sec/3600.0;
   if (dir=='W' || dir=='w') *d_lon *= -1.0;
}


/*****************************************************************************
   oil_to_deg
   Convert oilstock latitude and longitude to decimal degrees
   Inputs : lat - latitude character string
            lon - longitude character string
            d_lat - latitude
            d_lon - longitude
   Outputs : none
*****************************************************************************/
void oil_to_deg(char *lat, char *lon, double *d_lat, double *d_lon) {

   char    clat[16];
   char    clon[16];

   clat[0] = lat[0];
   clat[1] = lat[1];
   clat[2] = ':';
   clat[3] = lat[2];
   clat[4] = lat[3];
   clat[5] = ':';
   clat[6] = lat[4];
   clat[7] = lat[5];
   if (strlen(lat) == 7) {
      clat[8] = lat[6];
      clat[9] = 0;
   }
   else {
      clat[8] = '.';
      clat[9] = lat[7];
      clat[10] = lat[8];
      clat[11] = lat[9];
      clat[12] = 0;
   }

   clon[0] = lon[0];
   clon[1] = lon[1];
   clon[2] = lon[2];
   clon[3] = ':';
   clon[4] = lon[3];
   clon[5] = lon[4];
   clon[6] = ':';
   clon[7] = lon[5];
   clon[8] = lon[6];
   if (strlen(lon) == 8) {
      clon[9] = lon[7];
      clon[10] = 0;
   }
   else {
      clon[9] = '.';
      clon[10] = lon[8];
      clon[11] = lon[9];
      clon[12] = lon[10];
      clon[13] = 0;
   }
 
   dms_to_deg(clat,clon,d_lat,d_lon);
}


/*****************************************************************************
   xy_to_deg
   Convert xy to latitude longitude
   Inputs : x - x position (m)
            y - y position (m)
            lat - address to store latitude
            lon - address to store longitude
   Outputs : none
*****************************************************************************/
void xy_to_deg(double x, double y, double *lat, double *lon) {

   double term, xback, yback, zback;

   /* transform point to latitude/longitude */
   term = CONVERT_EarthRadius2 - x*x - y*y;
   term = term <= 0 ? 0 : sqrt( term );
   xback = -x * CONVERT_sin_ctrlon - y * CONVERT_sin_ctrlat * CONVERT_cos_ctrlon + term * CONVERT_cos_ctrlat * CONVERT_cos_ctrlon;
   yback =  x * CONVERT_cos_ctrlon - y * CONVERT_sin_ctrlat * CONVERT_sin_ctrlon + term * CONVERT_cos_ctrlat * CONVERT_sin_ctrlon;
   zback =  y * CONVERT_cos_ctrlat + term * CONVERT_sin_ctrlat;

   term = zback / CONVERT_EarthRadius;

   if (term <= -1)
      *lat = -90.0;
   else if (term >= 1)
      *lat = 90.0;
   else
      *lat = rad2deg(asin( term ));

   *lon = rad2deg(atan2( yback, xback ));
}


/*****************************************************************************
   deg_to_xy
   Convert Lat/Lon to xy
   Inputs : lon - longitude
            lat - latitude
            x - address to store x value (m)
            y - address to store y value (m)
   Outputs : none
*****************************************************************************/
void deg_to_xy(double lat, double lon, double *x, double *y) {

   double  placos;
   double  plasin;
   double  delong;

   lat = deg2rad(lat);
   lon = deg2rad(lon);
   placos = cos(lat) * CONVERT_EarthRadius;
   plasin = sin(lat) * CONVERT_EarthRadius;
   delong = lon - CONVERT_ctrlon;

   *x = placos * sin(delong);

   *y = CONVERT_cos_ctrlat * plasin - CONVERT_sin_ctrlat * placos * cos(delong);
}


/*****************************************************************************
   xy_to_deg
   Convert xy to latitude longitude
   Inputs : x - x position (m)
            y - y position (m)
            lat - address to store latitude
            lon - address to store longitude
   Outputs : none
*****************************************************************************/
void xy_to_rad(double x, double y, double *lat, double *lon) {

   double term, xback, yback, zback;

   /* transform point to latitude/longitude */
   term = CONVERT_EarthRadius2 - x*x - y*y;
   term = term <= 0 ? 0 : sqrt( term );
   xback = -x * CONVERT_sin_ctrlon - y * CONVERT_sin_ctrlat * CONVERT_cos_ctrlon + term * CONVERT_cos_ctrlat * CONVERT_cos_ctrlon;
   yback =  x * CONVERT_cos_ctrlon - y * CONVERT_sin_ctrlat * CONVERT_sin_ctrlon + term * CONVERT_cos_ctrlat * CONVERT_sin_ctrlon;
   zback =  y * CONVERT_cos_ctrlat + term * CONVERT_sin_ctrlat;

   term = zback / CONVERT_EarthRadius;

   if (term <= -1)
      *lat = -HALFPI;
   else if (term >= 1)
      *lat = HALFPI;
   else
      *lat = asin(term);

   *lon = atan2( yback, xback );
}


/*****************************************************************************
   deg_to_xy
   Convert Lat/Lon to xy
   Inputs : lon - longitude
            lat - latitude
            x - address to store x value (m)
            y - address to store y value (m)
   Outputs : none
*****************************************************************************/
void rad_to_xy(double lat, double lon, double *x, double *y) {

   double  placos;
   double  plasin;
   double  delong;

   placos = cos(lat) * CONVERT_EarthRadius;
   plasin = sin(lat) * CONVERT_EarthRadius;
   delong = lon - CONVERT_ctrlon;

   *x = placos * sin(delong);

   *y = CONVERT_cos_ctrlat * plasin - CONVERT_sin_ctrlat * placos * cos(delong);
}


/* **************************************************************
************************************************************** */
void init_earth(double center_lat, double center_lon, double earth_radius) {

   CONVERT_ctrlat = center_lat;	// Radians 
   CONVERT_ctrlon = center_lon;	// Radians
   CONVERT_EarthRadius = earth_radius;	// Meters 
   CONVERT_EarthRadius2 = CONVERT_EarthRadius * CONVERT_EarthRadius;

   /*
   *  The center of the scenario does not change during the course of an
   *  exercise.  Hence, precalculating these values here avoids the overhead
   *  of the trigonometric functions
   */
   CONVERT_sin_ctrlat = sin(CONVERT_ctrlat);
   CONVERT_cos_ctrlat = cos(CONVERT_ctrlat);
   CONVERT_sin_ctrlon = sin(CONVERT_ctrlon);
   CONVERT_cos_ctrlon = cos(CONVERT_ctrlon);

   CONVERT_sin_ctrlat_X_cos_ctrlon = CONVERT_sin_ctrlat * CONVERT_cos_ctrlon;
   CONVERT_sin_ctrlat_X_sin_ctrlon = CONVERT_sin_ctrlat * CONVERT_sin_ctrlon;
   CONVERT_cos_ctrlat_X_cos_ctrlon = CONVERT_cos_ctrlat * CONVERT_cos_ctrlon;
   CONVERT_cos_ctrlat_X_sin_ctrlon = CONVERT_cos_ctrlat * CONVERT_sin_ctrlon;
} // end init_scenario


